/*  ZIP plugin for Tux Commander
 *   version 0.3.8, designed for ZipArchive v3.0.2
 *  Copyright (C) 2007 Tomas Bzatek <tbzatek@users.sourceforge.net>
 *   Check for updates on tuxcmd.sourceforge.net
 *
 *  Uses ZipArchive library
 *   Copyright (C) 2000 - 2007 Artpol Software - Tadeusz Dracz
 *   http://www.artpol-software.com/ZipArchive/
 * 


 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string>
#include <stdlib.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <dirent.h>
#include <fnmatch.h>
#include <unistd.h>
#include <glib.h>

#include "vfs_types.h"
#include "strutils.h"

#include "ZipArchive.h"
#include "ZipPlatform.h"
#include "ZipCallback.h"



#define VERSION "0.3.8"
#define BUILD_DATE "2007-08-11"
#define DEFAULT_BLOCK_SIZE 65536


using namespace std;
extern "C" {


/******************************************************************************************************/
/**  Utilities                                                                                        */
/**************                                                                        ****************/                  

TVFSResult get_vfs_errorcode(int m_iCause)
{
  switch (m_iCause) {
    case 13: return cVFS_WriteErr;                              //  Permission denined
    case CZipException::noError: return cVFS_WriteErr;          //  No error.
    case CZipException::genericError: return cVFS_WriteErr;     //  An unknown error.
    case CZipException::badZipFile: return cVFS_ReadErr;        //  Damaged or not a zip file.
    case CZipException::badCrc: return cVFS_ReadErr;            //  Crc is mismatched.
    case CZipException::noCallback: return cVFS_Failed;         //  There is no spanned archive callback object set.
    case CZipException::aborted: return cVFS_Failed;            //  The disk change callback method returned false.
    case CZipException::abortedAction: return cVFS_Failed;      //  The action callback method returned false.
    case CZipException::abortedSafely: return cVFS_Failed;      //  The action callback method returned false, but the data is not corrupted.
    case CZipException::nonRemovable: return cVFS_WriteErr;     //  The device selected for the spanned archive is not removable.
    case CZipException::tooManyVolumes: return cVFS_WriteErr;   //  The limit of the maximum volumes reached.
    case CZipException::tooManyFiles: return cVFS_ReadErr;      //  The limit of the maximum files in an archive reached.
    case CZipException::tooLongData: return cVFS_ReadErr;       //  The filename, the comment or the local or central extra field of the file added to the archive is too long.
    case CZipException::tooBigSize: return cVFS_ReadErr;        //  The file size is too large to be supported.
    case CZipException::badPassword: return cVFS_Failed;        //  An incorrect password set for the file being decrypted.
    case CZipException::dirWithSize: return cVFS_ReadErr;       //  The directory with a non-zero size found while testing.
    case CZipException::internalError: return cVFS_WriteErr;    //  An internal error.
    case CZipException::notRemoved: return cVFS_WriteErr;       //  Error while removing a file
    case CZipException::notRenamed: return cVFS_WriteErr;       //  Error while renaming a file (under Windows call GetLastError() to find out more).
    case CZipException::platfNotSupp: return cVFS_WriteErr;     //  Cannot create a file for the specified platform.
    case CZipException::cdirNotFound: return cVFS_ReadErr;      //  The central directory was not found in the archive (or you were trying to open not the last disk of a segmented archive).
    case CZipException::noZip64: return cVFS_ReadErr;           //  The Zip64 format has not been enabled for the library, but is required to open the archive.
    case CZipException::noAES: return cVFS_ReadErr;             //  WinZip AES encryption has not been enabled for the library, but is required to decompress the archive.
    case CZipException::outOfBounds: return cVFS_Failed;        //  The collection is empty and the bounds do not exist.
    case CZipException::streamEnd: return cVFS_ReadErr;         //  Zlib library error.
    case CZipException::needDict: return cVFS_ReadErr;          //  Zlib library error.
    case CZipException::errNo: return cVFS_ReadErr;             //  Zlib library error.
    case CZipException::streamError: return cVFS_ReadErr;       //  Zlib library error.
    case CZipException::dataError: return cVFS_ReadErr;         //  Zlib library error.
    case CZipException::memError: return cVFS_ReadErr;          //  Zlib library or CZipMemFile error.
    case CZipException::bufError: return cVFS_ReadErr;          //  Zlib library error.
    case CZipException::versionError: return cVFS_ReadErr;      //  Zlib library error.
  }    
  return cVFS_Failed;  //  Default error
}



/******************************************************************************************************/
/**  Auxiliary classes                                                                                */
/**************                                                                        ****************/                  

//  Define the progress class and the class methods
struct ZIP_API CVFSZipActionCallback : public CZipActionCallback
{
        CVFSZipActionCallback()
        {
                m_uTotalToProcess = 0;
                m_uProcessed = 0;
                pCallBackProgress = NULL;
                data = NULL;
        }

        TVFSCopyCallBackFunc pCallBackProgress;
        void *data;

        virtual bool Callback(ZIP_SIZE_TYPE uProgress)
        {
          fprintf(stderr, "(II) Callback called, position = %lu; m_uTotalToProcess = %lu; m_uProcessed = %lu\n", 
                          uProgress, m_uTotalToProcess, m_uProcessed);
          bool ret = true;
          try {
            if (pCallBackProgress != NULL) ret = pCallBackProgress(m_uProcessed, m_uTotalToProcess, data);
          }
          catch (...) {
            fprintf(stderr, "(EE) extract_callback: Fatal error occured when calling pCallBackProgress\n");
          }
          return ret;
        }
};



//  Declaration of the global plugin object
struct TVFSGlobs {
  TVFSLogFunc log_func;
  int list_item_id;
  char *curr_dir;
  char *archive_path;

  CZipArchive *zip;
  CVFSZipActionCallback *extract_callback;

  int list_dir_index;
  bool archive_opened;
  unsigned long block_size;
  bool break_get_dir_size;
  bool archive_modified;

  GSList *file_list;
};




/***********************************************************************************************************************
*  Internal tree functions
********/

typedef struct {
  char *FileName;
  char *Path;
  char *FullPath;
  bool IsDir;
  long file_no;

  u_int64_t iSize;
  __time_t file_time;
  unsigned long permissions;
} TDirItem;


void recursive_get_listing(TVFSGlobs *globs, const char *start_dir)
{
  int iCount = globs->zip->GetCount();
  GSList *local_list = NULL;

  CZipFileHeader *fh;

  //  Prepare the current path include
  char *ANewPath = include_trailing_path_sep(start_dir);

  printf("Entering directory %s...\n", ANewPath);

  for (int i = 0; i < iCount; i++) {
    fh = new CZipFileHeader;
    if (globs->zip->GetFileInfo(*fh, i)) {
      //  Test to duplicate root entry
      if ((strlen(fh->GetFileName()) < 1) || (strcmp(fh->GetFileName(), "/") == 0)) {
        printf("$$$$$$ found duplicate root entry!!\n");
        continue;
      }
      //  the original filename
      char *s = include_leading_path_sep(fh->GetFileName());
      char *filename = exclude_trailing_path_sep(s);
      free(s);
//      fprintf(stderr, "filename = %s\n", filename);

      //  Test whether the file belongs to the current path
      char *curr_dir_part = strstr(filename, ANewPath);
      if (curr_dir_part == filename) {
        //  Strip the starting path
        char *curr_dir_name_strip = filename + strlen(ANewPath);
//        printf("$$$$$$$$$$$$ original = %s; curr_dir_name_strip = %s\n", filename, curr_dir_name_strip);


/*        //  Ensure that the first item is root
        if ((i == 0) && (strcmp(filename, "/") != 0)) {
          TDirItem *item = (TDirItem*)malloc(sizeof(TDirItem));
          item->Path = strdup("/");
          item->FileName = NULL;
          item->FullPath = strdup("/");
          item->IsDir = true;
          local_list = g_slist_append(local_list, item);
        }
*/

        //  Extract the current working directory or the filename in the current path
        char *curr_dir_name = NULL;
        char *curr_file_name = NULL;
        char *s = strstr(curr_dir_name_strip, "/");
        if (s != NULL) {
          int x = s - curr_dir_name_strip + 1;
          curr_dir_name = (char*)malloc(x);
          snprintf(curr_dir_name, x, "%s", curr_dir_name_strip);
          curr_file_name = strdup(s + 1);

          //  Check if the local list contains this directory. Otherwise, add it
          bool b = false;
          int max = g_slist_length(local_list);
          if (max > 0)
            for (unsigned int i = 0; i < g_slist_length(local_list); i++) {
              TDirItem *local_item = (TDirItem*)g_slist_nth_data(local_list, i);
              if ((strcmp(curr_dir_name, local_item->FileName) == 0) && (strcmp(ANewPath, local_item->Path) == 0))
              {
                b = true;
                break;
              }
            }
          if (!b) {
            printf("******** Adding %s / %s\n", ANewPath, curr_dir_name);
            TDirItem *item = (TDirItem*)malloc(sizeof(TDirItem));
            item->Path = strdup(ANewPath);
            item->FileName = strdup(curr_dir_name);
            int x = strlen(item->Path) + strlen(item->FileName) + 1;
            item->FullPath = (char*)malloc(x);
            snprintf(item->FullPath, x, "%s%s", item->Path, item->FileName);
            item->IsDir = true;
            item->iSize = 0;
            item->file_time = time(NULL);
            item->file_no = -1;
            item->permissions = S_IFDIR + S_IRUSR + S_IWUSR + S_IRGRP + S_IXGRP + S_IROTH + S_IXOTH;
            local_list = g_slist_append(local_list, item);
          }


        } else {
          //  This is the file in the current directory or it is a directory with no subdirectories
          curr_file_name = strdup(curr_dir_name_strip);

          //  Check if the local list contains this directory to not making more duplicates
          bool b = false;
          if (fh->IsDirectory()) {
            int max = g_slist_length(local_list);
            if (max > 0)
              for (unsigned int ii = 0; ii < g_slist_length(local_list); ii++) {
                TDirItem *local_item = (TDirItem*)g_slist_nth_data(local_list, ii);
                if ((strcmp(curr_file_name, local_item->FileName) == 0) && (strcmp(ANewPath, local_item->Path) == 0))
                {
                  local_item->file_time = (__time_t)fh->GetTime();
                  local_item->iSize = fh->m_uUncomprSize;
                  local_item->permissions = fh->GetSystemAttr() >> 16;
                  local_item->file_no = i;
                  printf("$$$$$$ duplicate entry %s\n", curr_file_name);
                  b = true;
                  break;
                }
              }
          }

          if (!b) {
            TDirItem *item = (TDirItem*)malloc(sizeof(TDirItem));
            item->Path = strdup(ANewPath);
            item->FileName = strdup(curr_file_name);
            int x = strlen(item->Path) + strlen(item->FileName) + 1;
            item->FullPath = (char*)malloc(x);
            snprintf(item->FullPath, x, "%s%s", item->Path, item->FileName);
            item->IsDir = fh->IsDirectory();
            item->iSize = fh->m_uUncomprSize;
            item->file_time = (__time_t)fh->GetTime();
            item->file_no = i;
            item->permissions = fh->GetSystemAttr() >> 16;
            local_list = g_slist_append(local_list, item);
          }
        }
//        printf("$$$$$$$$$$ curr_dir_name = %s; curr_file_name = %s\n", curr_dir_name, curr_file_name);
      }

      //  Free the objects and temporary strings
      free(filename);
      delete fh;
    }
  }


  //  printing the array
  for (unsigned int i = 0; i < g_slist_length(local_list); i++) {
    TDirItem *item = (TDirItem*)g_slist_nth_data(local_list, i);
    printf("  Adding %s %s (directory = %i)\n", item->Path, item->FileName, item->IsDir);
    globs->file_list = g_slist_append(globs->file_list, item);
  }

  //  do the recursion
  for (unsigned int i = 0; i < g_slist_length(local_list); i++) {
    TDirItem *item = (TDirItem*)g_slist_nth_data(local_list, i);
    if (item->IsDir) {
      int i = strlen(item->Path) + strlen(item->FileName) + 2;
      char *sx = (char*)malloc(i);
      snprintf(sx, i, "%s%s", item->Path, item->FileName);
      recursive_get_listing(globs, sx);
      free(sx);
    }
  }

/* //  Disabled due to reusing objects in globs->file_list
  //  free the array
  for (int i = 0; i < g_slist_length(local_list); i++) {
    TDirItem *item = (TDirItem*)g_slist_nth_data(local_list, i);
    if (item->FileName != NULL) free(item->FileName);
    if (item->Path != NULL) free(item->Path);
    free(item);
  }
  */
  g_slist_free(local_list);

  free(ANewPath);

  printf("Leaving directory %s...\n", start_dir);
}


void filelist_free(TVFSGlobs *globs);  //  Forward declaration

void build_global_filelist(TVFSGlobs *globs)
{
  int iCount = globs->zip->GetCount();

  //  Ensure the filelist is freed
  filelist_free(globs);

  //  list files in archive
  for (int i = 0; i < iCount; i++) {
    CZipFileHeader *fh = globs->zip->GetFileInfo(i);
    if (fh != NULL) 
      printf("  No: %i, %s, IsDir: %i, Size: %lu, SystemAttr = %lX, OriginalAttr = %lX\n", 
    	     i, (LPCTSTR)fh->GetFileName(), fh->IsDirectory(), fh->m_uUncomprSize, fh->GetSystemAttr(), fh->GetOriginalAttributes());
  }

  if (iCount > 0) recursive_get_listing(globs, "/");

  //  printing the array
  printf("\n\n\n\nPrinting the contents of the global filelist:\n\n");
  for (unsigned int i = 0; i < g_slist_length(globs->file_list); i++) {
    TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
    printf("  [%s]%s (directory = %i)\n", item->Path, item->FileName, item->IsDir);
  }
}

void filelist_free(TVFSGlobs *globs)
{
  try {
    if (g_slist_length(globs->file_list) > 0) {
      for (unsigned int i = 0; i < g_slist_length(globs->file_list); i++) {
        TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
        if (item->FileName != NULL) free(item->FileName);
        if (item->Path != NULL) free(item->Path);
        if (item->FullPath != NULL) free(item->FullPath);
        free(item);
      }
      g_slist_free(globs->file_list);
      globs->file_list = NULL;
    }
  }
  catch (...)  {
    printf ("(EE) filelist_free: Error freeing the file list\n");
  }
}


long filelist_find_file(TVFSGlobs *globs, bool FilesOnly, const char *directory, const char *filename = NULL)
{
  int count = g_slist_length(globs->file_list);
  if (count < 1) {
    printf("(EE) filelist_find_file: The archive is empty.\n");
    return -1;
  }

  bool SearchFullPath = filename == NULL;
  char *ADirectory = exclude_trailing_path_sep(directory);

  //  Try to retrieve file number
  bool found = false;
  long file_no = -1;
  if (count > 0) {
    for (int i = 0; i < count; i++) {
      TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
//      printf("filelist_find_file: '%s'\n", item->FullPath);
      if (SearchFullPath) {
        if ((strcmp(item->FullPath, ADirectory) == 0) && (item->file_no >= 0) && ((! FilesOnly) || (! item->IsDir))) {
          found = true;
          file_no = item->file_no;
//          printf("filelist_find_file: found = %ld\n", file_no);
          break;
        }
      } else {
        if ((strcmp(item->FileName, filename) == 0) && (strcmp(item->FullPath, ADirectory) == 0) && (item->file_no >= 0) && ((! FilesOnly) || (! item->IsDir))) {
          found = true;
          file_no = item->file_no;
          break;
        }
      }
    }
  }

  free(ADirectory);

  if ((! found) || (file_no < 0)) {
    printf("(EE) filelist_find_file: can't find the file specified: '%s/%s'\n", directory, filename);
    return -1;
  }
  return file_no;
}





////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////
//  Basic initialization functions
int VFSAllocNeeded()
{
  return sizeof(TVFSGlobs);
}

void VFSInit(TVFSGlobs *globs, TVFSLogFunc log_func)
{
  globs->archive_opened = false;
  globs->block_size = DEFAULT_BLOCK_SIZE;
  globs->archive_modified = false;

  globs->log_func = log_func;
  if (globs->log_func != NULL) globs->log_func("zip plugin: VFSInit");
}

void VFSDestroy(TVFSGlobs *globs)
{
  if (globs->log_func != NULL) globs->log_func("zip plugin: VFSDestroy");
}

int VFSVersion()
{
  return 2;
}

TVFSInfo VFSGetInfo()
{
  TVFSInfo module_info;
  module_info.Name = "ZIP plugin";
  module_info.Description = "ZIP archiving plugin";
  char *s = (char*)malloc(255);
  snprintf(s, 255, "version %s, build date: %s\nusing ZipArchive library v%s\n", VERSION, BUILD_DATE, CZipArchive::m_gszVersion);
  module_info.About = strdup(s);
  free(s);
  s = (char*)malloc(255);
  snprintf(s, 255, "Plugin Copyright (C) 2007 Tomáš Bžatek\n%s", CZipArchive::m_gszCopyright);
  module_info.Copyright = strdup(s);
  return module_info;
}

char *VFSGetPrefix(TVFSGlobs *globs)
{
//  return "zip";
  return globs->archive_path;
}

char *VFSGetExts()
{
  return "zip";
}


/**************************************************************************************************************************************/
/**************************************************************************************************************************************/

TVFSResult VFSOpen(TVFSGlobs *globs, char *sName)
{
  //  Initialize the objects
  globs->file_list = NULL;
  globs->curr_dir = NULL;
//  globs->zip = (CZipArchive*)malloc(sizeof(CZipArchive));    -- not needed?
  globs->zip = new CZipArchive;

  try
  {
    fprintf(stderr, "(--) VFSOpen: trying to open the file...\n");

    try
    {
      globs->zip->Open(sName, CZipArchive::zipOpen, 0);
    }
    catch (...)
    {
      printf("(!!) VFSOpen: error opening readwrite zip, trying readonly...\n");
      try
      {
        // try to open in read only mode (required if there's no write access on the media)
        globs->zip->Open(sName, CZipArchive::zipOpenReadOnly, 0);
      }
      catch (...)
      {
        printf("(EE) VFSOpen: error opening readonly zip\n");
        return cVFS_Failed;
      }
    }
    int iCount = globs->zip->GetCount(false);
    printf("(II) VFSOpen: %i records found, %i files.\n", iCount, globs->zip->GetCount(true));
    if (iCount < 1) return cVFS_Failed;

    //  Build the global file list
    build_global_filelist(globs);

    //  Set the progress callback
    globs->extract_callback = new CVFSZipActionCallback;
    globs->zip->SetCallback(globs->extract_callback, CZipActionCallback::cbExtract);
    globs->zip->SetCallback(globs->extract_callback, CZipActionCallback::cbAdd);

    //  Set automatic flushing of changes to disk
    globs->zip->SetAutoFlush(true);
  }
  catch (CZipException e)
  {
    printf ("(EE) VFSOpen: Error while processing archive %s\n%s\n", (LPCTSTR) sName, (LPCTSTR)e.GetErrorDescription());
    if (e.m_szFileName.IsEmpty()) printf("\n");
    else printf("(EE) VFSOpen: Filename in error object: %s\n\n", (LPCTSTR)e.m_szFileName);
    globs->zip->Close(true);
    return cVFS_Failed;
  }
  catch (...)
  {
    printf ("(EE) VFSOpen: Unknown error while processing archive %s\n\n", (LPCTSTR) sName);
    globs->zip->Close(true);
    return cVFS_Failed;
  }

  //  Everything goes OK
  globs->archive_path = strdup(sName);
  globs->archive_modified = false;
  return cVFS_OK;
}


TVFSResult VFSClose(TVFSGlobs *globs)
{
  //  Closing the archive...
  fprintf(stderr, "(II) VFSClose: Closing the archive...\n");
  try {
	if (globs->archive_modified) globs->zip->Flush();
    globs->zip->Close(CZipArchive::afNoException, false);
    //*** In case of inconsistency, try using afWriteDir value
    //    (Use when an exception was thrown. The Close method writes the 
    //     central directory structure to the archive, so that the archive should be usable.)
  }
  catch (CZipException e) {
    fprintf(stderr, "(EE) VFSClose: Error while closing archive: %s\n", (LPCTSTR)e.GetErrorDescription());
    return cVFS_Failed;
  }


  //  Freeing the ZIP objects...
  fprintf(stderr, "(II) VFSClose: Freeing ZipArchive objects...\n");
  try {
    delete globs->extract_callback;
//    free(globs->extract_callback);       --  crashing
//    globs->zip->~CZipArchive();          --  crashing
    delete globs->zip;
//    free(globs->zip);                    --  crashing
  }
  catch (...)  {
    fprintf(stderr, "(EE) VFSClose: Error freeing ZipArchive objects\n");
    return cVFS_Failed;
  }


  //  Freeing the filelist
  fprintf(stderr, "(II) VFSClose: Freeing filelist...\n");
  filelist_free(globs);

  //  Free the rest...
  free(globs->archive_path);

  return cVFS_OK;
}

char *VFSGetPath(TVFSGlobs *globs)
{
  return globs->curr_dir;
}

u_int64_t VFSGetFileSystemFree(TVFSGlobs *globs, char *APath)
{
  return 0;
}

u_int64_t VFSGetFileSystemSize(TVFSGlobs *globs, char *APath)
{
  return globs->zip->GetOccupiedSpace();
}




///////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TVFSResult VFSChangeDir(TVFSGlobs *globs, char *NewPath)
{
  if (NewPath == NULL) {
    printf("(EE) VFSChangeDir: NewPath is NULL!\n");
    return cVFS_Failed;
  }

  printf ("(--) VFSChangeDir: Going to change dir from %s\n", NewPath);
  char *ANewPath = exclude_trailing_path_sep(NewPath);
  if (strlen(ANewPath) <= 0) ANewPath = strdup("/");
  printf ("(--) VFSChangeDir: Going to change dir to   %s\n", ANewPath);

  //  Find the directory in the filelist
  int x = g_slist_length(globs->file_list);
  bool found = false;
  if (strcmp(ANewPath, "/") == 0) found = true; else
  if (x > 0) {
    for (unsigned int i = 0; i < g_slist_length(globs->file_list); i++) {
      TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
      if ((item->IsDir) && (strcmp(item->FullPath, ANewPath) == 0)) {
        found = true;
        break;
      }
    }
  }

  printf ("(II) VFSChangeDir: Found = %i\n", found);
  if (! found) {
    printf ("(EE) VFSChangeDir: Directory '%s' not found.\n", ANewPath);
    return cVFS_Failed;
  }

  if (globs->curr_dir != NULL) free(globs->curr_dir);
  globs->curr_dir = strdup(ANewPath);

  free(ANewPath);
  return cVFS_OK;
}

int VFSLogin(TVFSGlobs *globs, char *user, char *pass)
{
  printf ("(II) VFSLogin: Going to set the password...\n");
  try {
    globs->zip->SetPassword(pass);
    return cVFS_OK;
  }
  catch (...)
  {
    printf ("(EE) VFSLogin: Changing password failed. Closed archive maybe?\n");
    return cVFS_Failed;
  }
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

void build_file_info(TDirItem *dir_item, TVFSItem *Item, bool include_full_path)
{
  if (! include_full_path) Item->sFileName = strdup(dir_item->FileName);
  else Item->sFileName = strdup(dir_item->FullPath);
  Item->iSize = dir_item->iSize;
  Item->sLinkTo = NULL;
  Item->m_time = dir_item->file_time;
  Item->a_time = dir_item->file_time;
  Item->c_time = dir_item->file_time;
  printf("  [%s]%s (directory = %i), (time [%ld] = %s), permissions = %lX\n", dir_item->Path, dir_item->FileName, dir_item->IsDir, dir_item->file_time, ctime(&dir_item->file_time), dir_item->permissions);
  if (dir_item->IsDir) {
    Item->ItemType = vDirectory;
    Item->iMode = dir_item->permissions | S_IRUSR | S_IXUSR;
  } else {
    Item->ItemType = vRegular;
    Item->iMode = dir_item->permissions | S_IRUSR;
  }
}


//  Returns true if item found
bool perform_listing(TVFSGlobs *globs, TVFSItem *Item, char *list_path)
{
  int count = g_slist_length(globs->file_list);
  if (globs->list_dir_index >= count) return false;

  //  Prepare the current path include
  char *ANewPath;
  ANewPath = include_trailing_path_sep(list_path);
  printf("(--) process_filelist: Working dir: %s\n", ANewPath);

  bool found = false;

  for (int i = globs->list_dir_index; i < count; i++)
  {
    globs->list_dir_index = i;

    TDirItem *dir_item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
    if (strcmp(dir_item->Path, ANewPath) == 0) {
      found = true;
      build_file_info(dir_item, Item, false);
      break;
    }
  }

  free(ANewPath);

  return found;
}


TVFSResult VFSListFirst(TVFSGlobs *globs, char *sDir, TVFSItem *Item)
{
  globs->list_dir_index = 0;

  if (perform_listing(globs, Item, sDir)){
    printf("(II) VFSListFirst: found file: %s\n\n", Item->sFileName);
    return cVFS_OK;
  } else {
    printf("(II) VFSListFirst: no more files\n\n");
    return cVFS_No_More_Files;
  }
}

TVFSResult VFSListNext(TVFSGlobs *globs, char *sDir, TVFSItem *Item)
{
  globs->list_dir_index++;

  if (perform_listing(globs, Item, sDir)){
    printf("(II) VFSListNext: found file: %s\n\n", Item->sFileName);
    return cVFS_OK;
  } else {
    printf("(II) VFSListNext: no more files\n\n");
    return cVFS_No_More_Files;
  }
}

TVFSResult VFSListClose(TVFSGlobs *globs)
{
  return cVFS_OK;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////
long VFSFileExists(TVFSGlobs *globs, const char *FileName, const long Use_lstat)
{
  int count = g_slist_length(globs->file_list);
  if (count < 1) return false;

  char *ANewPath;
  ANewPath = exclude_trailing_path_sep(FileName);
  bool found = false;

  if (strcmp(ANewPath, "/") == 0) found = true; else
  if (count > 0) {
    for (int i = 0; i < count; i++) {
      TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
      if (strcmp(item->FullPath, ANewPath) == 0) {
        found = true;
        break;
      }
    }
  }

  free(ANewPath);
  printf("(II) VFSFileExists(%s) = %i\n", FileName, found);

  return found;
}

TVFSResult VFSFileInfo(TVFSGlobs *globs, char *AFileName, TVFSItem *Item)
{
  printf("(II) VFSFileInfo: getting file info for '%s'\n", AFileName);

  int count = g_slist_length(globs->file_list);
  if (count < 1) return false;

  char *ANewPath;
  ANewPath = exclude_trailing_path_sep(AFileName);
  bool found = false;

  if (strcmp(ANewPath, "/") == 0) {
    //  We have to manually fill some info for the root item
    found = true;
    Item->sFileName = strdup("/");
    Item->iSize = 0;
    Item->sLinkTo = NULL;
    Item->m_time = 0;
    Item->a_time = 0;
    Item->c_time = 0;
    Item->ItemType = vDirectory;
    Item->iMode = S_IRWXO + S_IRWXG + S_IRWXU;
  } else
  if (count > 0) {
    for (int i = 0; i < count; i++) {
      TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
      if (strcmp(item->FullPath, ANewPath) == 0) {
        found = true;
        build_file_info(item, Item, true);
        break;
      }
    }
  }

  free(ANewPath);
  if (found) return cVFS_OK;
  else {
    printf("(EE) VFSFileInfo(%s): specified file not found\n", AFileName);
    return cVFS_Failed;
  }
}

/******************************************************************************************************/
/**  Recursive tree size counting                                                                     */
/**************                                                                        ****************/                  

u_int64_t VFSGetDirSize(TVFSGlobs *globs, char *APath)
{
  if (globs == NULL) return 0;
  u_int64_t Size = 0;
  globs->break_get_dir_size = 0;

  int count = g_slist_length(globs->file_list);
  if (count < 1) return 0;
  char *ANewPath = include_trailing_path_sep(APath);

  //  Main search loop
  for (int i = 0; i < count; i++) {
    if (globs->break_get_dir_size) break;
    TDirItem *item = (TDirItem*)g_slist_nth_data(globs->file_list, i);
    if ((strstr(item->Path, ANewPath) == item->Path) && (! item->IsDir))
      Size += item->iSize;
  }

  free(ANewPath);
  globs->break_get_dir_size = 0;
  return Size;
}


void VFSBreakGetDirSize(TVFSGlobs *globs)
{
  printf("(WW) VFSBreakGetDirSize: calling break\n");
  if (globs != NULL) globs->break_get_dir_size = true;
}


/******************************************************************************************************/
/**  Methods modifying the archive                                                                    */
/**************                                                                        ****************/                  

TVFSResult VFSMkDir(TVFSGlobs *globs, const char *sDirName)
{
  if ((sDirName == NULL) || (strlen(sDirName) < 1)) {
    printf("(EE) VFSMkDir: The value of 'sDirName' is NULL or empty\n");
	return cVFS_Failed;
  }
  if ((strlen(sDirName) < 1) || (strcmp(sDirName, "/") == 0)) {
    printf("(EE) VFSMkDir: Invalid value '%s' (duplicate root entry?)\n", sDirName);
    return cVFS_Failed;
  }
  printf ("(II) VFSMkDir: Going to create new directory '%s'...\n", sDirName);

  try {
	try {  
      CZipFileHeader header;
      globs->zip->SetFileHeaderAttr(header, 0x41ED0010);  //  alternatively use ZipPlatform::GetDefaultAttributes();
      char *s = exclude_leading_path_sep(sDirName);
      header.SetFileName(s);
      free(s);
      header.SetTime(time(NULL));
      bool bRet = globs->zip->OpenNewFile(header, 0, NULL);
      globs->zip->CloseNewFile();
      if (! bRet) {
        printf("(EE) VFSMkDir: Error creating new directory '%s'\n", sDirName);
        return cVFS_Failed;
      } 
      globs->archive_modified = true;	
      build_global_filelist(globs);
      return cVFS_OK;
	}
    catch (CZipException e) {
      globs->zip->CloseNewFile(true);
      fprintf(stderr, "(EE) VFSMkDir: Error creating new directory '%s': [%d] %s, archive closed = %d.\n", 
                      sDirName, e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }
  }
  catch (...)
  {
    printf("(EE) VFSMkDir: Error creating new directory '%s'\n", sDirName);
    return cVFS_Failed;
  }
}

TVFSResult VFSRemove(TVFSGlobs *globs, const char *APath)
{
  printf("(II) VFSRemove: Going to remove the file '%s'...\n", APath);

  char *AFile = exclude_trailing_path_sep(APath);
  long file_no = filelist_find_file(globs, false, AFile);
  free(AFile);

  if (file_no < 0) {
    printf("(EE) VFSRemove: can't find the file specified: '%s'\n", APath);
    return cVFS_Failed;
  }

  try {
    try {
	  if (! globs->zip->RemoveFile(file_no)) {
        printf("(EE) VFSRemove: Delete file '%s' failed.\n", APath);
        return cVFS_Failed;
      }
      build_global_filelist(globs);
      globs->archive_modified = true;
      printf("(II) VFSRemove OK.\n");
      
      //  Test for the sparse ZIP central directory
      char *AFile1 = exclude_trailing_path_sep(APath);
      char *AFile2 = g_path_get_dirname(AFile1);
      char *AFile3 = exclude_trailing_path_sep(AFile2);
      if ((strlen(AFile3) > 0) && (strcmp(AFile3, "/") != 0)) {
        printf("(II) VFSRemove: AFile1: '%s', AFile2: '%s', AFile3: '%s'\n", AFile1, AFile2, AFile3);
        file_no = filelist_find_file(globs, false, AFile2);
        printf("(II) VFSRemove: deleted: '%s', parent: '%s', file_no = %ld\n", APath, AFile3, file_no);
        if (file_no < 0) {
          printf("(WW) VFSRemove: sparse ZIP archive detected, adding empty directory: '%s'\n", AFile3);
          VFSMkDir(globs, AFile3);
        }
      }  
      free(AFile1);   free(AFile2);   free(AFile3);
      
      return cVFS_OK;
    }
    catch (CZipException e) {
      fprintf(stderr, "(EE) VFSRemove: Delete file '%s' failed: [%d] %s, archive closed = %d.\n", 
                       APath, e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }
  }
  catch (...)
  {
    printf("(EE) VFSRemove: Delete file '%s' failed.\n", APath);
    return cVFS_Failed;
  }
}

TVFSResult VFSRename(TVFSGlobs *globs, const char *sSrcName, const char *sDstName)
{
  printf ("(II) VFSRename: Going to rename/move the file '%s' to '%s'...\n", sSrcName, sDstName);

  char *AFile = exclude_trailing_path_sep(sSrcName);
  char *ADestFile = exclude_trailing_path_sep(sDstName);
  long file_no = filelist_find_file(globs, false, AFile);
  free(AFile);

  if (file_no < 0) {
    printf("(EE) VFSRename: can't find the file specified: '%s'\n", sSrcName);
    return cVFS_Failed;
  }

  try {
    try {
      if (! globs->zip->RenameFile(file_no, ADestFile)) {
        printf ("(EE) VFSRename: Rename/move file '%s' failed.\n", sSrcName);
        return cVFS_Failed;
      }
      free(ADestFile);
      build_global_filelist(globs);
      globs->archive_modified = true;	
      return cVFS_OK;
    }
    catch (CZipException e) {
      fprintf(stderr, "(EE) VFSRename: Rename/move file '%s' failed: [%d] %s, archive closed = %d.\n",
                      sSrcName, e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }
  }
  catch (...)
  {
    printf ("(EE) VFSRename: Rename/move file failed.\n");
    return cVFS_Failed;
  }
}

TVFSResult VFSMakeSymLink(TVFSGlobs *globs, const char *NewFileName, const char *PointTo)
{
  fprintf(stderr, "(EE) VFSMakeSymLink: Symbolic links not supported in ZIP archives.\n");
  return cVFS_Not_Supported;
}

TVFSResult VFSChmod(TVFSGlobs *globs, const char *FileName, const uint Mode)
{
  printf("(II) VFSChmod: Going to change permissions of the file '%s'...\n", FileName);

  char *AFile = exclude_trailing_path_sep(FileName);
  long file_no = filelist_find_file(globs, false, AFile);
  free(AFile);

  if (file_no < 0) {
    printf("(EE) VFSChmod: can't find the file specified: '%s'\n", FileName);
    return cVFS_Failed;
  }

  try {
    try {
      //  Change the header data
      CZipFileHeader *header = globs->zip->GetFileInfo(file_no);
      if (header == NULL) {
        printf("(EE) VFSChmod: Permissions modification of the file '%s' failed: NULL returned by GetFileInfo()\n", FileName);
        return cVFS_Failed;
      }
      //  We need to change only 0xF000FFFF mask
      //  The 0xF_______ bits represents file/directory type, the 0x____FFFF represents ZIP attributes and 0x_FFF____ represents unix permissions
      printf("(II) VFSChmod: Current permissions: %lX, stripped: %lX, setting to: %X, modified: %lX\n", 
              header->GetSystemAttr(), header->GetSystemAttr() & 0xF000FFFF, Mode & 0xFFF, (header->GetSystemAttr() & 0xF000FFFF) + ((Mode & 0xFFF) << 16));
      globs->zip->SetFileHeaderAttr(*header, (header->GetSystemAttr() & 0xF000FFFF) + ((Mode & 0xFFF) << 16));
      globs->zip->RemoveCentralDirectoryFromDisk();

      printf("(II) VFSChmod OK.\n");
      build_global_filelist(globs); 
      globs->archive_modified = true;
      return cVFS_OK;
    }
    catch (CZipException e) {
      globs->zip->CloseNewFile(true);
      fprintf(stderr, "(EE) VFSChmod: DateTime modification of the file '%s' failed: [%d] %s, archive closed = %d.\n", 
                       FileName, e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }
  }
  catch (...)
  {
    printf("(EE) VFSChmod: DateTime modification of the file '%s' failed.\n", FileName);
    return cVFS_Failed;
  }
}

TVFSResult VFSChown(TVFSGlobs *globs, const char *FileName, const uint UID, const uint GID)
{
  fprintf(stderr, "(EE) VFSChown: Owner changing is not supported in ZIP archives.\n");
  return cVFS_Not_Supported;
}

//  The following function uses silly workaround
TVFSResult VFSChangeTimes(TVFSGlobs *globs, char *APath, long mtime, long atime)
{
  printf ("(II) VFSChangeTimes: Going to change date/times of the file '%s'...\n", APath);

  char *AFile = exclude_trailing_path_sep(APath);
  long file_no = filelist_find_file(globs, false, AFile);
  free(AFile);

  if (file_no < 0) {
    printf("(EE) VFSChangeTimes: can't find the file specified: '%s'\n", APath);
    return cVFS_Failed;
  }

  try {
    try {
      //  Change the header data
      CZipFileHeader *header = globs->zip->GetFileInfo(file_no);
      if (header == NULL) {
        printf("(EE) VFSChangeTimes: DateTime modification of the file '%s' failed: NULL returned by GetFileInfo()\n", APath);
        return cVFS_Failed;
      }
      header->SetTime(mtime);
      globs->zip->RemoveCentralDirectoryFromDisk();
     
      printf("(II) VFSChangeTimes OK.\n");
      build_global_filelist(globs); 
      globs->archive_modified = true;
      return cVFS_OK;
    }
    catch (CZipException e) {
      globs->zip->CloseNewFile(true);
      fprintf(stderr, "(EE) VFSChangeTimes: DateTime modification of the file '%s' failed: [%d] %s, archive closed = %d.\n", 
                       APath, e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }
  }
  catch (...)
  {
    printf("(EE) VFSChangeTimes: DateTime modification of the file '%s' failed.\n", APath);
    return cVFS_Failed;
  }
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

TVFSFileDes VFSOpenFile(TVFSGlobs *globs, const char *APath, int Mode, int *Error)
{
  *Error = cVFS_Not_Supported;
  return (TVFSFileDes)0;
}

TVFSResult VFSCloseFile(TVFSGlobs *globs, TVFSFileDes FileDescriptor)
{
  return cVFS_Not_Supported;
}

u_int64_t VFSFileSeek(TVFSGlobs *globs, TVFSFileDes FileDescriptor, u_int64_t AbsoluteOffset, int *Error)
{
  *Error = cVFS_Not_Supported;
  return 0;
}

int VFSReadFile(TVFSGlobs *globs, TVFSFileDes FileDescriptor, void *Buffer, int ABlockSize, int *Error)
{
  *Error = cVFS_Not_Supported;
  return 0;
}

int VFSWriteFile(TVFSGlobs *globs, TVFSFileDes FileDescriptor, void *Buffer, int BytesCount, int *Error)
{
  *Error = cVFS_Not_Supported;
  return 0;
}

void VFSSetBlockSize(TVFSGlobs *globs, int Value)
{
  globs->block_size = Value;
}

int VFSIsOnSameFS(TVFSGlobs *globs, const char *Path1, const char *Path2)
{
  printf("(II) VFSIsOnSameFS: Not supported in ZIP archives.\n");
  return true;
}

int VFSTwoSameFiles(TVFSGlobs *globs, const char *Path1, const char *Path2)
{
  printf("(II) VFSTwoSameFiles: Not supported in ZIP archives, comparing by paths.\n");
  char *f1 = exclude_trailing_path_sep(Path1);
  char *f2 = exclude_trailing_path_sep(Path2);
  bool b = strcmp(f1, f2) == 0;
  free(f1);
  free(f2);
  return b;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////


//  Known issues: - crashes when no space left on NFS mounts, probably unhandled exception in further ZipArchive code (repro: Gentoo, Ubuntu)
TVFSResult VFSCopyOut(TVFSGlobs *globs, const char *sSrcName, const char *sDstName, TVFSCopyCallBackFunc pCallBackProgress, void *data, int Append)
{
  if ((sSrcName == NULL)  || (sDstName == NULL) || (strlen(sSrcName) < 1) || (strlen(sDstName) < 1)) {
    printf("(EE) VFSCopyOut: The value of 'sSrcName' or 'sDstName' is NULL or empty\n");
    return cVFS_Failed;
  }
  
  printf("(II) VFSCopyOut: copying file '%s' out to '%s'\n", sSrcName, sDstName);

  long file_no = filelist_find_file(globs, true, sSrcName);
  if (file_no < 0) {
    printf("(EE) VFSCopyOut: can't find source file '%s'\n", sSrcName);
    return cVFS_ReadErr;
  }

  char *s = exclude_trailing_path_sep(sDstName);
  char *dest_path = extract_file_path(s);
  char *dest_filename = extract_file_name(s);
  free(s);

  //  Set callback data
  globs->extract_callback->data = data;
  globs->extract_callback->pCallBackProgress = pCallBackProgress;

  //  Perform extract
  try {
    try {
      if (! globs->zip->ExtractFile(file_no, dest_path, false, dest_filename, globs->block_size)) {
        globs->zip->CloseFile(NULL, true);
        fprintf(stderr, "(EE) VFSCopyOut: Error while copying out, archive closed = %d.\n", globs->zip->IsClosed());
        return cVFS_WriteErr;
      }
      fprintf(stderr, "(II) VFSCopyOut: copy OK, archive closed = %d.\n", globs->zip->IsClosed());
    }
    catch (CZipException e) {
      globs->zip->CloseFile(NULL, true);
      fprintf(stderr, "(EE) VFSCopyOut: Error while copying out: [%d] %s, archive closed = %d.\n", 
    		          e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }  
  }
  catch (...)
  {
    fprintf(stderr, "(EE) VFSCopyOut: Fatal error while copying out..., archive closed = %d.\n", globs->zip->IsClosed());
    return cVFS_Failed;
  }  

  free(dest_path);
  free(dest_filename);
  return cVFS_OK;
}

//  Known issues: - archive corruption when no space left on device
TVFSResult VFSCopyIn(TVFSGlobs *globs, const char *sSrcName, const char *sDstName, TVFSCopyCallBackFunc pCallBackProgress, void *data, int Append)
{
  if ((sSrcName == NULL)  || (sDstName == NULL) || (strlen(sSrcName) < 1) || (strlen(sDstName) < 1)) {
    printf("(EE) VFSCopyIn: The value of 'sSrcName' or 'sDstName' is NULL or empty\n");
    return cVFS_Failed;
  }

  printf("(II) VFSCopyIn: copying file '%s' in to '%s'\n", sSrcName, sDstName);

  //  Set callback data
  globs->extract_callback->data = data;
  globs->extract_callback->pCallBackProgress = pCallBackProgress;

  try {
    try {
      char *s = exclude_leading_path_sep(sDstName);
	  if (! globs->zip->AddNewFile(sSrcName, s, -1, CZipArchive::zipsmSafeSmart, globs->block_size)) {
        globs->zip->CloseNewFile(true);
        globs->zip->CloseFile(NULL, true);
        build_global_filelist(globs);
        fprintf(stderr, "(EE) VFSCopyIn: Error while copying in, archive closed = %d.\n", globs->zip->IsClosed());
        return cVFS_WriteErr;
      }
	  free(s);

	  globs->zip->Flush();
      printf("(II) VFSCopyIn: copy OK, archive closed = %d.\n", globs->zip->IsClosed());
      build_global_filelist(globs);
      globs->archive_modified = true;	
    }
    catch (CZipException e) {
      globs->zip->CloseNewFile(true);
      globs->zip->CloseFile(NULL, true);
      build_global_filelist(globs);
      fprintf(stderr, "(EE) VFSCopyIn: Error while copying in: [%d] %s, archive closed = %d.\n", 
                      e.m_iCause, (LPCTSTR)e.GetErrorDescription(), globs->zip->IsClosed());
      return get_vfs_errorcode(e.m_iCause);
    }
  }
  catch (...)
  {
    fprintf(stderr, "(EE) VFSCopyIn: Fatal error while copying in..., archive closed = %d.\n", globs->zip->IsClosed());
    return cVFS_Failed;
  }

  return cVFS_OK;
}


///////////////////////////////
//  end of extern "C"
}




/***
* Todo:
* 
* - look for faster array classes in GLIB and modify all listing functions (idea: use g_node_ tree storage)
* - implement an "opened" flag - is it really needed? + add tests to all functions
* - readonly checking IsReadOnly() + add tests to all functions modifying the archive
* - after VFS API update implement archive testing, compression level setting, retrieving compression info for each file and archive, readonly flags, begin-action, end-action (no refresh, faster operations) ...
*
*/



